#!/usr/local/bin/perl

use Getopt::Std;
use Net::IP;

use strict;

my $usage = "usage:$0 [-h] inet6num\n";
sub Help() {
    print <<EOP;
$usage
    h - this message
EOP
}

my %opts;
unless (getopts("h",\%opts)) {
    print STDERR $usage;
    exit -1;
}
if (defined $opts{h}) {
    Help();
    exit(0);
}

my @info;
ParseWhoisEvents();

my @sorted = sort byprefix @info;
=cut
    print "DEBUG: sorted prefixes:\n";
    map { print "\t", $_->{inet}->print(), "\n" } @sorted;
=cut

RemoveCounted();
#PrintRootPrefixes();
CountRootPrefixes();
#Count();


sub ParseWhoisEvents {
    my $linenum = 0;
my $last_rir;
my $key =0;
    while (<>) {
        next if /^\s*$/;

	my @values = split /,/;
	foreach my $value (@values) { $value =~ s/^\s+//; $value=~ s/\s+$//;}

	my @prefixes = @values;
	foreach my $prefix (@prefixes) {
	    my $inet = new Net::IP($prefix);
	    unless (defined $inet) {
	      print STDERR "ERROR: line $linenum: malformed prefix '$prefix'\n";
	      next;
	    }

	    # size as positive or negative number of bits difference from 24;
	    # negative means longer than /24; positive means shorter than 24;
	    # 1 / (2 ** $size) is the block size in terms of a single /24
            my $length = $inet->prefixlen();
	    my $size = 24 - $length;

	    my $info = {
		"inet" => $inet,
		"size" => $size,
		"root_prefix" => 1,
		};
	    push @info, $info;
	}
	$linenum++;
    }
}

sub RemoveCounted {
    my @stack;
    foreach my $info (@sorted) {
	my $inet = $info->{inet};
	while ($#stack >= 0
	       && $inet->overlaps($stack[0]{inet}) == $IP_NO_OVERLAP) {
#print "DEBUG: popping ", $stack[0]{inet}->print(), " which doesn't overlap ", $inet->print(), "\n";
	    shift @stack;
	}

	foreach my $inet_on_stack (map { $_->{inet} } @stack) {
	  unless (within_prefix($inet_on_stack, $inet)) {
	    print STDERR "ERROR: ", $inet_on_stack->print(), " doesn't completely enclose ", $inet->print(), "\n";
	  }
	}

	if ($#stack >= 0) {
	  print "".(scalar @stack)." OVERLAP: ", $inet->print(), "\n\t",
	    join("\n\t", map { $_->{inet}->print() } @stack), "\n"
	      if @stack >= 3;

	    $info->{root_prefix} = 0;
	    if ($stack[0]{date} <= $info->{date}) {
		$info->{size_exclude} .= " " . $info->{size};
	    } else {
		$stack[0]{size_exclude} .= " " . $info->{size};
	    }
	}
	unshift @stack, $info;
    }
}


sub byprefix {
  my $a_inet = $a->{inet};
  my $b_inet = $b->{inet};

  # bincomp() only compares the first address in each prefix so we
  # must separately compare the prefix length to order prefixes
  # from largest to smallest (e.g., 2001::/16 before 2001::/32).
  #
  # bincomp() doesn't support "eq".

  return -1 if $a_inet->bincomp('lt', $b_inet);
  return 1 if $a_inet->bincomp('gt', $b_inet);
  return $a_inet->prefixlen() <=> $b_inet->prefixlen();
}


sub PrintRootPrefixes {
  foreach my $info (grep { $_->{root_prefix} } @sorted) {
    print "ROOT: ", $info->{inet}->print(), "\n";
  }
}


sub CountRootPrefixes {
  my %root_counts;
  foreach my $info (grep { $_->{root_prefix} } @sorted) {
    ++$root_counts{$info->{size}};
  }

  my @lengths = sort {$a<=>$b} keys %root_counts;
  my $min_length = $lengths[0];

  foreach my $k (@lengths) {
    print $k, " (/", 24 - $k, ")\t", $root_counts{$k}, "\n";
  }

  print "\n", ("=" x 70), "\n";

  my %counts_norm = %root_counts;

  for (my $k = $min_length; $k <= 31; $k++) {
    next unless defined $counts_norm{$k};

    my $v = $counts_norm{$k};
    my $vdiv2 = int($v / 2);
    my $vmod2 = $v % 2;
    $counts_norm{$k + 1} += $vdiv2;
    $counts_norm{$k} = $vmod2;
  }

  foreach my $k (sort {$a<=>$b} keys %counts_norm) {
    my $v = $counts_norm{$k};
    next if $v == 0;
    print $k, " (/", 24 - $k, ")\t$v\n";
  }

  my $total_slash24s = 0;
  my $total_slash8s = 0;
  foreach my $k (sort {$a<=>$b} keys %counts_norm) {
    my $v = $counts_norm{$k};
    next if $v == 0;

    $total_slash24s += $v * 2 ** $k;
    $total_slash8s += $v * 2 ** ($k - 16);
  }

  print "(root prefixes) total /24-equivalent coverage: $total_slash24s\n";
  print "(root prefixes) total /8-equivalent coverage: $total_slash8s\n";

  my $equiv_prefix_length = 8 - (log($total_slash8s) / log(2));
  print "(root prefixes) equivalent prefix length: $equiv_prefix_length\n";
}


sub within_prefix
{
  my ($pref_a, $pref_b) = @_;

  my $result = $pref_a->overlaps($pref_b);
  return undef unless defined $result;
  return ($result == $IP_B_IN_A_OVERLAP || $result == $IP_IDENTICAL);
}
